import yaml
from yaml.scanner import ScannerError


class OpenStackResource:
    def __init__(self, client, os_cloud="admin"):
        self.client = client
        self.os_cloud = os_cloud

    def _exec(self, method, *args, yaml_output=True):
        """
        :param method: operation on resource i.e "list", "create"
        :param args: list of string positional args should go first i.e.:
            ["--flavor=m1.tiny"]
            ["vm-test-0", "--flavor=m1.tiny"]
        """
        command = ['env', 'OS_CLOUD=' + self.os_cloud, 'openstack', self.resource, method, *args]

        if yaml_output:
            command += ['-f', 'yaml']
            result = self.client.exec(command)
            try:
                return yaml.safe_load(result)
            except ScannerError:
                raise Exception(f"Response YAML deserialization error. Plain response:\n{result}")

        return self.client.exec(command)

    def list(self, args, yaml_output=True):
        return self._exec('list', *args, yaml_output=yaml_output)

    def create(self, args, yaml_output=True):
        return self._exec('create', *args, yaml_output=yaml_output)

    def delete(self, args, yaml_output=False):
        return self._exec('delete', *args, yaml_output=yaml_output)

    def show(self, args):
        return self._exec('show', *args)

    def set(self, args):
        return self._exec('set', *args, yaml_output=False)

    def add(self, args, yaml_output=False):
        return self._exec('add', *args, yaml_output=yaml_output)

    def check(self, args, yaml_output=False):
        return self._exec('check', *args, yaml_output=yaml_output)

    @property
    def resource(self):
        return self._resource


class Aggregate(OpenStackResource):
    _resource = 'aggregate'

    def add_host(self, *args):
        return self._exec('add host', *args, yaml_output=False)

    def remove_host(self, *args):
        return self._exec('remove host', *args, yaml_output=False)


class LoadBalancer(OpenStackResource):
    _resource = 'loadbalancer'

    def pool_list(self, args, yaml_output=True):
        return self._exec('pool list',
                          yaml_output=yaml_output,
                          *args)


class ZoneBlacklist(OpenStackResource):
    _resource = 'zone blacklist'


class Role(OpenStackResource):
    _resource = 'role'

    def remove(self, args, yaml_output=False):
        return self._exec('remove', *args, yaml_output=yaml_output)


class User(OpenStackResource):
    _resource = 'user'


class Image(OpenStackResource):
    _resource = 'image'


class Server(OpenStackResource):
    _resource = 'server'


class Flavor(OpenStackResource):
    _resource = 'flavor'


class Stack(OpenStackResource):
    _resource = 'stack'

    def resource_list(self, args, yaml_output=False):
        return self._exec('resource list',
                          yaml_output=yaml_output,
                          *args)

    def event_list(self, args, yaml_output=False):
        return self._exec('event list',
                          yaml_output=yaml_output,
                          *args)

    def resource_show(self, args, yaml_output=True):
        return self._exec('resource show',
                          yaml_output=yaml_output,
                          *args)

    def output_show(self, args, yaml_output=True):
        return self._exec('output show',
                          yaml_output=yaml_output,
                          *args)


class ResourceProvider(OpenStackResource):
    _resource = "resource provider"


class ResourceProviderUsage(OpenStackResource):
    _resource = "resource provider usage"


class Network(OpenStackResource):
    _resource = 'network'


class NetworkAgent(OpenStackResource):
    _resource = 'network agent'


class Host(OpenStackResource):
    _resource = 'host'


class Notification(OpenStackResource):
    _resource = 'notification'


class Hypervisor(OpenStackResource):
    _resource = 'hypervisor'


class ComputeService(OpenStackResource):
    _resource = 'compute service'


class Subnet(OpenStackResource):
    _resource = 'subnet'


class Segment(OpenStackResource):
    _resource = 'segment'

    def host_create(self, args, yaml_output=False):
        return self._exec('host create',
                          yaml_output=yaml_output,
                          *args)

    def host_update(self, args, yaml_output=False):
        return self._exec('host update',
                          yaml_output=yaml_output,
                          *args)

    def host_list(self, args, yaml_output=True):
        return self._exec('host list',
                          yaml_output=yaml_output,
                          *args)


class Volume(OpenStackResource):
    _resource = 'volume'


class VolumeService(OpenStackResource):
    _resource = 'volume service'


class Secret(OpenStackResource):
    _resource = 'secret'


class Quota(OpenStackResource):
    _resource = 'quota'


class Console(OpenStackResource):
    _resource = 'console'

    def log_show(self, args):
        return self._exec('log show',
                          yaml_output=False,
                          *args)


class Port(OpenStackResource):
    _resource = 'port'


class Floatingip(OpenStackResource):
    _resource = 'floating ip'


class Loadbalancer(OpenStackResource):
    _resource = 'loadbalancer'

    def amphora_list(self, args, yaml_output=False):
        return self._exec('amphora list',
                          yaml_output=yaml_output,
                          *args)

    def amphora_show(self, args, yaml_output=True):
        return self._exec('amphora show',
                          yaml_output=yaml_output,
                          *args)


class Baremetal(OpenStackResource):
    _resource = 'baremetal'

    def node_create(self, args, yaml_output=False):
        return self._exec('node create',
                          yaml_output=yaml_output,
                          *args)

    def node_delete(self, args, yaml_output=False):
        return self._exec('node delete',
                          yaml_output=yaml_output,
                          *args)

    def node_list(self, args, yaml_output=True):
        return self._exec('node list',
                          yaml_output=yaml_output,
                          *args)

    def node_manage(self, args, yaml_output=False):
        return self._exec('node manage',
                          yaml_output=yaml_output,
                          *args)

    def node_provide(self, args, yaml_output=False):
        return self._exec('node provide',
                          yaml_output=yaml_output,
                          *args)
