import pytest

import yaml
import json
import time
import os

from numpy import percentile
from kubernetes.client import V1Namespace

from si_tests import logger
from si_tests import settings
from si_tests.fixtures.kubectl import create_secret
from si_tests.utils import templates, scale_lab_trend_client


LOG = logger.logger
RALLY_LABEL = {settings.RALLY_LABEL_KEY: settings.RALLY_LABEL_VALUE}


def run_tests(secret_name, kubectl_client, name=settings.RALLY_POD_NAME):
    """Run rally tests"""
    ns = settings.RALLY_NAMESPACE

    options = {
        'RALLY_POD_NAME': name,
        'KUBECONFIG_SECRET_NAME': secret_name
    }

    if not settings.RALLY_ON_DEDICATED_CLUSTER:
        template = templates.render_template(
            settings.RALLY_POD_YAML_WITHOUT_DEDICATED_CLUSTER, options)
    else:
        template = templates.render_template(
            settings.RALLY_POD_YAML, options)
    LOG.debug(template)
    json_body = json.dumps(yaml.load(template, Loader=yaml.SafeLoader))

    if kubectl_client.pods.present(name=name, namespace=ns):
        LOG.warning("Pod '{0}' already exists. Deleting..".format(name))
        # kubectl_client.api_core.delete_namespaced_pod(name, ns)
        kubectl_client.pods.get(name=name, namespace=ns).delete()

    pod = kubectl_client.pods.create(name=name,
                                     namespace=ns,
                                     body=json.loads(json_body))
    pod.wait_phase('Running', timeout=300, interval=10)
    return pod


def wait_test_results(pod, timeout=settings.RALLY_WAIT_TIMEOUT, container=None):
    pod.wait_test(filepath='./test_*', timeout=timeout, container=container)


def check_test_results(pod, container=None):
    """Check if the file ./test_passed presents in the pod

    Return True if present, False otherwise
    """
    command = ['/bin/sh', '-c', 'test -f ./test_passed && echo "OK"']
    return pod.exec(command, container=container) == "OK\n"


def download_test_results(pod, compression=None, container=None):
    pod.cp_from_pod(source_dir='/rally-reports', compression=compression, container=container)


def _get_rally_node(nodesList):
    for node in nodesList:
        if node.exists_labels(RALLY_LABEL):
            return node
    return nodesList[0]


def add_label_daemonset_node(kubectl_client):
    label_selector_daemonset = ('stacklight!=enabled,!node-role.kubernetes.io/master,' +
                                '{0}!={1}'.format(settings.RALLY_LABEL_KEY, settings.RALLY_LABEL_VALUE))
    nodes_daemonset = kubectl_client.nodes.list(label_selector=label_selector_daemonset)
    label_daemonset = {'rally': 'daemonset'}
    for node in nodes_daemonset:
        node.add_label(label_daemonset)


def prepare_rally_node(kubectl_client, node=None):
    taint = {"key": settings.RALLY_TOLERATION_KEY,
             "value": settings.RALLY_TOLERATION_VALUE,
             "effect": settings.RALLY_TOLERATION_EFFECT}
    if node:
        rally_node = kubectl_client.nodes.get(name=node)
        LOG.info('Pre-selected node {}'.format(rally_node.get_name()))
    else:
        label_selector = 'stacklight!=enabled,!node-role.kubernetes.io/master'
        nodes = kubectl_client.nodes.list(label_selector=label_selector)
        rally_node = _get_rally_node(nodes)
        LOG.info('Selected node {}'.format(rally_node.get_name()))
    if not rally_node.exists_labels(RALLY_LABEL):
        rally_node.add_label(RALLY_LABEL)
    if not rally_node.exists_taint(taint):
        rally_node.add_taint(taint)


def create_kubeconfig_secret(kubectl_client, content, namespace, name):
    config = ''
    if settings.INTERNAL_K8S_ENDPOINT_ENABLED:
        INTERNAL_K8S_ENDPOINT = 'https://kubernetes.default.svc.cluster.local:443'
        kubeconfig = yaml.load(content, Loader=yaml.SafeLoader)
        for cluster in kubeconfig['clusters']:
            cluster['cluster']['server'] = INTERNAL_K8S_ENDPOINT
        config = yaml.dump(kubeconfig)
    else:
        config = content
    create_secret(kubectl_client, config, namespace, name)


@pytest.mark.parametrize("_", ["CLUSTER_NAME={0}"
                               .format(settings.TARGET_CLUSTER)])
def test_k8s_rally(kaas_manager, _):
    env_type = settings.RALLY_ENV_TYPE
    env_name = settings.RALLY_ENV_NAME
    upload_report = settings.RALLY_UPLOAD_SCALE_LAB_TRENDS

    # Collect data and prepare for test
    if settings.RALLY_ON_DEDICATED_CLUSTER:
        # cluster with rally:
        # 1. get testcluster k8s client
        ts_name = settings.TESTCLUSTER_NAME
        ts_ns = kaas_manager.get_namespace(settings.TESTCLUSTER_NAMESPACE)
        ts_cluster = ts_ns.get_cluster(ts_name)
        ts_client = ts_cluster.k8sclient
        # target cluster:
        # 2. get target (tested) cluster kubeconfig
        tg_name = settings.TARGET_CLUSTER
        tg_ns = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)
        tg_cluster = tg_ns.get_cluster(tg_name)
        _, tg_kubeconfig = tg_cluster.get_kubeconfig_from_secret()
    else:
        ts_name = tg_name = settings.TARGET_CLUSTER
        ts_ns = tg_ns = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)
        ts_cluster = tg_cluster = ts_ns.get_cluster(ts_name)
        ts_client = ts_cluster.k8sclient
        _, tg_kubeconfig = tg_cluster.get_kubeconfig_from_secret()
    # 3. create ns in target cluster
    rally_ns = settings.RALLY_NAMESPACE
    if not ts_client.namespaces.present(name=rally_ns):
        LOG.info("Creating Namespace '{0}'".format(rally_ns))
        ns_body = V1Namespace(api_version='v1',
                              kind='Namespace',
                              metadata={'name': rally_ns})
        ts_client.api_core.create_namespace(body=ns_body)
    else:
        LOG.info("Namespace '{0}' already exists".format(rally_ns))
    # 4. create secret in target cluster
    LOG.debug("Creating target cluster kubeconfig secret")
    secret_name = '-'.join(('kubeconfig', tg_name))
    if ts_client.secrets.present(name=secret_name, namespace=rally_ns):
        LOG.warning("Secret '{0}' already exists. Deleting.."
                    .format(secret_name))
        ts_client.api_core.delete_namespaced_secret(
            secret_name, rally_ns
        )
    create_kubeconfig_secret(ts_client, tg_kubeconfig, rally_ns, secret_name)

    # Run test
    pod_name = container_name = settings.RALLY_POD_NAME
    if not settings.RALLY_ON_DEDICATED_CLUSTER:
        LOG.info('Adding label and taint for rally node')
        prepare_rally_node(ts_client, settings.RALLY_NODE_NAME)
    LOG.info('Adding labels for the daemonset scenario')
    add_label_daemonset_node(ts_client)
    LOG.info("Starting rally pod")
    pod = run_tests(secret_name, ts_client, name=pod_name)
    LOG.info("Waiting for results ...")
    start = time.time()
    wait_test_results(pod=pod,
                      timeout=settings.RALLY_WAIT_TIMEOUT,
                      container=container_name)
    end = time.time()

    download_test_results(pod, compression='-z', container=container_name)
    k8s_rally_pass = check_test_results(pod, container=container_name)

    # Get steal time during the rally test
    if settings.RALLY_GET_STEAL_TIME:
        LOG.info('Get steal time from control nodes')
        control_nodes = [x.get_k8s_node_name() for x in tg_cluster.get_machines(machine_type="control")]
        result = {}
        steal_time_values_per_node = {x: [] for x in control_nodes}
        LOG.debug(f"Managment nodes: {control_nodes}")
        query = f"rate(node_cpu_seconds_total{{mode='steal',node=~'{'|'.join(control_nodes)}'}}[1m])"
        LOG.debug(query)
        prometheus_response = ts_cluster.prometheusclient.get_query_range(query=query, start=start, end=end, step=60)
        LOG.debug(prometheus_response)

        # Prepare steal time metrics
        for metrics in prometheus_response:
            if metrics['metric'].get('node'):
                for value in metrics['values']:
                    steal_time_values_per_node[metrics['metric']['node']].append(float(value[1]))

        # Get 95 percentile
        LOG.info('Steal time on nodes')
        for node_name, list_of_values in steal_time_values_per_node.items():
            steal_time_percentile = percentile(list_of_values, 95)
            result[node_name] = steal_time_percentile
            LOG.info(f"{node_name}: {steal_time_percentile}")

        with open(os.path.join(settings.ARTIFACTS_DIR, 'steal_time.json'), 'w') as f:
            json.dump(result, f)

    # Checking all pods are in consistent state after test run
    tg_cluster.check.check_k8s_pods()

    # Uploading rally results to scale lab trends
    if upload_report:
        sltc = scale_lab_trend_client.ScaleLabTrendClient(
            kaas_manager=tg_cluster._manager,
            kaas_manager_mgmt=kaas_manager
        )
        sltc.uploadReport(
            env_type=env_type,
            env_name=env_name,
            path=os.path.join(settings.ARTIFACTS_DIR, 'rally-report.json')
        )

    if settings.RALLY_CHECK_RESULTS:
        assert k8s_rally_pass,  "Some of k8s rally tests failed"
